package de.tu_darmstadt.tk.SmartHomeNetworkSim.control;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Observer;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Protocol;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;

/**
 * Controller which allows interaction with the Model and Simulation
 *
 * @author Andreas T. Meyer-Berg
 */
public class Controller {
	/**
	 * {@link Model} which stores the smart home model
	 */
	private Model model;
	
	/**
	 * Controller for Import of user defined java classes
	 */
	private ImportController importController;
	
	/**
	 * Controller which allows configuration of the program
	 */
	private ConfigurationController configurationController;
	
	/**
	 * Create a new Controller, which controls the given model and allows
	 * manipulation of this model
	 * 
	 * @param model
	 *            model, that should be controlled
	 */
	public Controller(Model model) {
		this.model = model;
		configurationController = new ConfigurationController(this.model, this);
		importController = new ImportController(this.model, this);
	}
	
	/**
	 * Returns the ImportController, which manages the user imported classes, 
	 * which can be used in the framework
	 * @return {@link ImportController}
	 */
	public ImportController getControllerImport(){
		return importController;
	}
	
	/**
	 * Returns the controller responsible for configuration of the program
	 * @return {@link ConfigurationController}
	 */
	public ConfigurationController getControllerConfiguration(){
		return configurationController;
	}
	
	/**
	 * Moves the SmartDevice device to the specified location, if it exists
	 * 
	 * @param device
	 *            the device to move
	 * @param x
	 *            new x position
	 * @param y
	 *            new y position
	 * @param z
	 *            new z position
	 */
	public void moveSmartDevice(SmartDevice device, int x, int y, int z) {
		device.setX(x);
		device.setY(y);
		device.setZ(z);

	}

	/**
	 * Changes the dimension of the model, without scaling the SmartDevice
	 * positions
	 * 
	 * @param width
	 *            the new width to set
	 * @param height
	 *            the new height to set
	 * @param depth
	 *            the new depth to set
	 */
	public void setDimension(int width, int height, int depth) {
		setDimension(width, height, depth, false);
	}

	/**
	 * Changes the dimension of the model and moves SmartDevices to new relative
	 * positions if {@code stretchModel} is true
	 * 
	 * @param width
	 *            the new width to set
	 * @param height
	 *            the new height to set
	 * @param depth
	 *            the new depth to set
	 * @param stretchModel
	 *            true, if smartDevice positions should be stretched
	 */
	public void setDimension(int width, int height, int depth, boolean stretchModel) {
		// Don't allow to small Models
		if (width < 400)
			width = 400;
		if (height < 400)
			height = 400;
		if (depth < 400)
			depth = 400;

		if (stretchModel) {
			/**
			 * Factor that changes the x position. Example: 1 -> positions stay
			 * the same 0.5 -> x position is halved 2 -> x position is doubled
			 */
			double width_factor = ((double) width) / ((double) model.getWidth());
			/**
			 * Factor that changes the y position. Example: 1 -> positions stay
			 * the same 0.5 -> y position is halved 2 -> y position is doubled
			 */
			double height_factor = ((double) height) / ((double) model.getHeight());
			/**
			 * Factor that changes the z position. Example: 1 -> positions stay
			 * the same 0.5 -> z position is halved 2 -> z position is doubled
			 */
			double depth_factor = ((double) depth) / ((double) model.getDepth());
			// Update all device positions
			for (SmartDevice d : model.getDevices()) {
				if (width_factor != 1.0)
					d.setX(scalePos(d.getX(), width_factor, width));
				if (height_factor != 1.0)
					d.setY(scalePos(d.getY(), height_factor, height));
				if (depth_factor != 1.0)
					d.setZ(scalePos(d.getZ(), depth_factor, depth));
			}
		}
		model.setWidth(width);
		model.setHeight(height);
		model.setDepth(depth);
	}

	/**
	 * Calculates the new scaled position, which should be in Bounds:
	 * {@code radius < newPosition <
	 * (upperBound - radius)}
	 * 
	 * @param oldPosition
	 *            previous position that should be scaled
	 * @param factor
	 *            how much it is stretched/shrunk
	 * @param upperBound
	 *            upper bound of the frame, which should not be exceeded
	 * @return new position that was calculated
	 */
	private int scalePos(int oldPosition, double factor, int upperBound) {
		/**
		 * New Position that should be validated and returned
		 */
		int newPosition = (int) Math.round(factor * oldPosition);
		// Update if it moves out of the frame
		if (newPosition < getDevice_visualization_radius())
			newPosition = getDevice_visualization_radius();
		if (newPosition >= upperBound - getDevice_visualization_radius())
			newPosition = upperBound - getDevice_visualization_radius();

		return newPosition;
	}

	/**
	 * Adds SmartDevice to the Model and verifies that it is inside the model
	 * bounds
	 * 
	 * @param sd
	 *            SmartDevice which should be added to the model
	 */
	public void addSmartDevice(SmartDevice sd) {
		if (model.getDevices().contains(sd))
			return;
		model.addDevices(sd);
		// validate Position
		sd.setX(scalePos(sd.getX(), 1.0, model.getWidth()));
		sd.setY(scalePos(sd.getY(), 1.0, model.getHeight()));
		sd.setZ(scalePos(sd.getZ(), 1.0, model.getDepth()));
	}

	/**
	 * Creates a new SmartDevice at the given Position. The Device is moved
	 * into the model bounds, if the position is outside the bounds
	 * 
	 * @param name
	 *            name of the smartDevice
	 * @param x_pos
	 *            position on the x-Axis
	 * @param y_pos
	 *            position on the y-Axis
	 * @param z_pos
	 *            position on the z-Axis
	 */
	public void createSmartDevice(String name, int x_pos, int y_pos, int z_pos) {
		SmartDevice sd = new SmartDevice(name);
		sd.setX(scalePos(x_pos, 1.0, model.getWidth()));
		sd.setY(scalePos(y_pos, 1.0, model.getHeight()));
		sd.setZ(scalePos(z_pos, 1.0, model.getDepth()));
		model.addDevices(sd);
	}

	/**
	 * Deletes the given SmartDevice toDelete, removes it from its links and
	 * connections.
	 * 
	 * @param toDelete
	 *            the SmartDevice to delete
	 */
	public void deleteSmartDevice(SmartDevice toDelete) {
		if (toDelete == null)
			return;
		// Remove from Links
		for (Link link : toDelete.getLinks())
			removeSmartDeviceFromLink(toDelete, link);
		// Remove Links from Device
		toDelete.getLinks().clear();

		// Delete Connections
		for (Port p : toDelete.getPorts()) {
			//Skip ports that are not connected
			if(p.getConnection()==null)continue;
			/**
			 * Connection of the current port
			 */
			Connection c = p.getConnection();
			
			//remove from protocol
			if(c.getProtocol()!=null)
				c.getProtocol().removeDevice(p);
			
			//remove from connection
			c.removeSmartDevice(p);
			if (c.getStatus()==Connection.FINISHED||c.getStatus()==Connection.TERMINATED) {
				//if connection terminated - remove rest of ports
				for (Port sd : c.getParticipants()) {
					if (sd != null && sd != p)
						sd.getOwner().removePort(sd);
				}
			}
		}
		//Remove all ports from the device
		toDelete.getPorts().clear();
		model.getDevices().remove(toDelete);
	}

	/**
	 * Removes the SmartDevice from the given Link
	 * 
	 * @param toRemove the Device that should be removed from the link
	 * @param link the Link, which contains the SmartDevice
	 */
	public void removeSmartDeviceFromLink(SmartDevice toRemove, Link link) {
		link.removeDevice(toRemove);
	}
	
	/**
	 * Notifies Observer, which may update their visualization/state
	 */
	public void notifyObservers(){
		model.setChanged();
		model.notifyObservers();
	}
	
	/**
	 * Adds Observer to the model, which will be notified, when the model changes
	 * 
	 * @param observer observer to be to notified on changes
	 */
	public void addObserver(Observer observer){
		model.addObserver(observer);
	}
	
	/**
	 * Removes Observer from the model. The Observer will no longer be notified on changes 
	 * 
	 * @param observer observer to be removed
	 */
	public void removeObserver(Observer observer){
		model.deleteObserver(observer);
	}
	
	/**
	 * @return the width
	 */
	public int getWidth() {
		return model.getWidth();
	}

	/**
	 * @param width the width to set
	 */
	public void setWidth(int width) {
		model.setWidth(width);
	}

	/**
	 * @return the height
	 */
	public int getHeight() {
		return model.getHeight();
	}

	/**
	 * @param height the height to set
	 */
	public void setHeight(int height) {
		model.setHeight(height);
	}

	/**
	 * @return the depth
	 */
	public int getDepth() {
		return model.getDepth();
	}

	/**
	 * @param depth the depth to set
	 */
	public void setDepth(int depth) {
		model.setDepth(depth);
	}
	
	/**
	 * @return the device_visualization_radius
	 */
	private int getDevice_visualization_radius() {
		return configurationController.getDeviceVisualizationRadius();
	}
	
	/**
	 * Adds Link to the model
	 * @param link link to add
	 */
	public void addLink(Link link){
		model.addConnectionNetwork(link);
	}
	
	/**
	 * Removes Link from the Model
	 * @param link link to remove
	 */
	public void removeLink(Link link){
		model.getConnectionNetworks().remove(link);
	}
	
	public Collection<Link> getLinks(){
		return model.getConnectionNetworks();
	}
	
	public void addConnection(Connection connection){
		model.addConnection(connection);
	}
	
	public void removeConnection(Connection connection){
		model.getConnections().remove(connection);
	}
	
	public Collection<Connection> getConnections(){
		return model.getConnections();
	}

	public void addLinkToDevice(Link newLink, SmartDevice smartDevice) {
		if(!newLink.getDevices().contains(smartDevice)){
			newLink.addDevice(smartDevice);
			smartDevice.addLink(newLink);
		}
		
	}

	public void removeLinkFromDevice(Link newLink, SmartDevice smartDevice) {
		newLink.removeDevice(smartDevice);
		smartDevice.removeLink(newLink);
	}
	
	/**
	 * Changes Roles of the device to the newRole, returns true, if the role was successfully changed, false if not.
	 * If false is returned, the device will no longer be part of this connection.
	 * 
	 * @param protocol protocol which should be edited
	 * @param device device which should be added
	 * @param newRole new role of the device
	 * @return true if new role was set, false if not
	 */
	public boolean changeRoleOfDevice(Protocol protocol, Connection con, Port device, int newRole){
		if(newRole < 0 || newRole >= protocol.getNumberOfRoles()){
			protocol.removeDevice(device);
			if(con.getParticipants().contains(device))
				con.removeSmartDevice(device);
			if(device.getConnection()!=null)
				device.setConnection(null);
			return false;
		} else if(protocol.getDevicesWithRole(newRole).contains(device)){
			if(!con.getParticipants().contains(device))
				con.addSmartDevice(device);
			if(device.getConnection()!=con)
				device.setConnection(con);
			return true;
		}
		else{
			protocol.removeDevice(device);
			boolean added = protocol.addDeviceOfRole(device, newRole);
			if(added){
				if(!con.getParticipants().contains(device))
					con.addSmartDevice(device);
				if(device.getConnection()!=con)
					device.setConnection(con);
				return true;
			}else{
				if(con.getParticipants().contains(device))
					con.removeSmartDevice(device);
				if(device.getConnection()!=null)
					device.setConnection(null);
				return false;
			}
		}
	}
	
	public void removeDeviceFromConnection(Port p, Connection connection){
		connection.removeSmartDevice(p);
		connection.getProtocol().removeDevice(p);
		p.setConnection(null);
	}
	
	/**
	 * Adds Device p to the connection with the specified role
	 * @param p Port/Device to be added
	 * @param connection Connection
	 * @param role new role of the device (See {@link Protocol} Roles)
	 * @return true, if the device was added
	 */
	public boolean addDeviceToConnection(Port p, Connection connection, int role){
		if(connection.getProtocol().getDevicesWithRole(role).contains(p) || connection.getProtocol().addDeviceOfRole(p, role)){
			//If port already has the role, or it could be assigned - just check the fields
			if(!connection.getParticipants().contains(p))
				connection.addSmartDevice(p);
			if(p.getConnection()!=connection)
				p.setConnection(connection);
			return true;
		}else {
			//Device could not be added -> Remove
			if(p.getConnection()==connection)
				p.setConnection(null);
			if(connection.getParticipants().contains(p))
				connection.removeSmartDevice(p);
			if(connection.getProtocol().getDevices().contains(p))
				connection.getProtocol().removeDevice(p);
			return false;
		}
	}

	/**
	 * Changes the link of the given connection to the new link. Returns true if it was successfully changed, false if it could not be changed.
	 * @param connection
	 * @param newLink
	 * @return true on successful change, false on failure & restore
	 */
	public boolean changeLinkOfConnection(Connection connection, Link newLink) {
		// TODO Auto-generated method stub
		if(newLink != null){
			connection.getLink().removeConnection(connection);
			connection.setLink(newLink);
			for(Port p:connection.getParticipants()){
				if(p.getOwner()!=null&& !newLink.getDevices().contains(p.getOwner())){
					this.addLinkToDevice(newLink, p.getOwner());
				}
			}
			connection.setLink(newLink);
			newLink.addConnection(connection);
			return true;
		}else{
			return false;
		}
	}
	
	public Connection changeConnectionType(Connection connection, Class<? extends Connection> newConnectionClass) {
		
		Connection newCon = null;
		try{
			newCon = newConnectionClass.newInstance();
			newCon.setProtocol(connection.getProtocol());
			newCon.setLink(connection.getLink());
			connection.getLink().addConnection(newCon);
			newCon.setStatus(connection.getStatus());
			newCon.setPacketLossProbability(connection.getPacketLossProbability());
			newCon.setName(connection.getName());
			for (Iterator<Port> p = connection.getParticipants().iterator(); p.hasNext();) {
				Port type = (Port) p.next();
				newCon.addSmartDevice(type);
				type.setConnection(newCon);
				
			}
		}catch(Exception e){
			System.out.println("Error while changing protocol: "+e.toString());
			/**
			 * Restore old connection on Failure
			 */
			if(newCon != null){
				newCon.setProtocol(null);
				if(newCon.getLink()!=null)
					newCon.getLink().removeConnection(newCon);
				newCon.setLink(null);
				for (Iterator<Port> p = connection.getParticipants().iterator(); p.hasNext();) {
					Port type = (Port) p.next();
					if(type.getConnection()!=connection)
						type.setConnection(connection);
					newCon.removeSmartDevice(type);
					
				}
				
				
			}
			return null;
		}
		if(this.getConnections().contains(connection)){
			this.removeConnection(connection);
		}
		this.addConnection(newCon);	
		getControllerConfiguration().getConfigurationManager().getSelectionModel().clickedConnection.clear();
			
			
		connection.setLink(null);
		connection.setProtocol(null);
		connection.setStatus(Connection.TERMINATED);
		connection.setName("deleted");
		for(Port p: newCon.getParticipants()){
			connection.removeSmartDevice(p);			
		}
		
		notifyObservers();
		
		return newCon;
	}
	
	/**
	 * Deletes the network model, removes all Devices, Connections and Links
	 */
	public void deleteNetworkModel(){
		/**
		 * Devices which should be deleted
		 */
		LinkedList<SmartDevice> devicesToDelete = new LinkedList<SmartDevice>(model.getDevices());
		for(SmartDevice d: devicesToDelete)
			deleteSmartDevice(d);
		devicesToDelete.clear();
		/**
		 * Connections which should be deleted
		 */
		LinkedList<Connection> connectionsToDelete = new LinkedList<Connection>(model.getConnections());
		for(Connection c: connectionsToDelete)
			deleteConnection(c);
		connectionsToDelete.clear();
		/**
		 * Links which should be delted
		 */
		LinkedList<Link> linksToDelete = new LinkedList<Link>(model.getConnectionNetworks());
		for(Link l: model.getConnectionNetworks())
			deleteLink(l);
		linksToDelete.clear();
		/**
		 * Update the GUI
		 */
		notifyObservers();
	}

	/**
	 * Deletes the Connection c and all references
	 * @param c Connection to be deleted
	 */
	public void deleteConnection(Connection c) {
		if(c == null)return;
		for(Port p:c.getParticipants())
			removeDeviceFromConnection(p, c);
		removeConnectionFromLink(c, c.getLink());
		c.setStatus(Connection.TERMINATED);
		removeConnection(c);
	}

	/**
	 * Deletes Link l and all references
	 * @param l Link to be deleted
	 */
	public void deleteLink(Link l) {
		if(l==null)return;
		for(SmartDevice d : l.getDevices())
			removeLinkFromDevice(l, d);
		for(Connection c:l.getConnections())
			removeConnectionFromLink(c,l);
		l.getPackets().clear();
		removeLink(l);
	}
	
	/**
	 * Removes Connection from Link
	 * @param c Connection to be removed
	 * @param l Link to be removed
	 */
	private void removeConnectionFromLink(Connection c, Link l) {
		if(c!=null && c.getLink()==l){
			c.setLink(null);
		}
		if(l!=null){
			l.removeConnection(c);		
		}
	}
}
